/* $Id: client-db.c,v 1.9 2001/08/20 01:04:24 robey Exp $
 *
 * Copyright (C) 2001  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <glib.h>
#include "client.h"
#include "client-db.h"

static const DBModule *module;


void db_set_module (const DBModule *mod)
{
        module = mod;
}

void *
db_init (const GData *params)
{
        return (*(module->init)) (params);
}

void
db_shutdown (void *state)
{
        return (*(module->shutdown)) (state);
}

gboolean
db_add_client (void *state, const Client *client)
{
        return (*(module->add_client)) (state, client);
}

gboolean
db_update_login (void *state, const Client *client)
{
        return (*(module->update_login)) (state, client);
}

gboolean
db_update_logout (void *state, const Client *client)
{
        return (*(module->update_logout)) (state, client);
}

gboolean
db_update_client (void *state, const Client *client)
{
        return (*(module->update_client)) (state, client);
}

gboolean
db_lookup_client (void *state, const char *id, Client *client)
{
        return (*(module->lookup_client)) (state, id, client);
}

gboolean
db_check_password (void *state, const char *id, const char *password, int flags)
{
        return (*(module->check_password)) (state, id, password, flags);
}

gboolean
db_set_password (void *state, const Client *client, const char *password)
{
        return (*(module->set_password)) (state, client, password);
}

gboolean
db_get_seen (void *state, const char *id, time_t *last_login, time_t *last_logout, int flags)
{
        return (*(module->get_seen)) (state, id, last_login, last_logout, flags);
}

gboolean
db_get_messages (void *state, const Client *client, GList **messages)
{
	return (*(module->get_messages))(state, client, messages);
}

gboolean
db_add_message (void *state, const char *id, const char *message)
{
	return (*(module->add_message))(state, id, message);
}

void
db_free_messages (GList *messages)
{
        g_list_foreach(messages, (GFunc)g_free, NULL);
        g_list_free(messages);
}

gboolean
db_get_buddies (void *state, Client *client)
{
        return (*(module->get_buddies))(state, client);
}

gboolean
db_update_buddy (void *state, const Client *client, const Buddy *buddy)
{
        return (*(module->update_buddy))(state, client, buddy);
}

gboolean
db_add_buddy (void *state, const Client *client, const Buddy *buddy)
{
        return (*(module->add_buddy))(state, client, buddy);
}

gboolean
db_delete_buddy (void *state, const Client *client, const Buddy *buddy)
{
        return (*(module->delete_buddy))(state, client, buddy);
}
