/* $Id: client-db.h,v 1.8 2001/08/20 01:04:24 robey Exp $
 *
 * Copyright (C) 2001  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef _CLIENT_DB_H_
#define _CLIENT_DB_H_

#include <glib.h>


typedef struct {
        void * (*init) (const GData *params);
        void (*shutdown) (void *state);
        gboolean (*add_client) (void *state, const Client *client);
        gboolean (*update_login) (void *state, const Client *client);
        gboolean (*update_logout) (void *state, const Client *client);
        gboolean (*update_client) (void *state, const Client *client);
        gboolean (*lookup_client) (void *state, const char *id, Client *client);
        gboolean (*check_password) (void *state, const char *id, const char *password, int flags);
        gboolean (*set_password) (void *state, const Client *client, const char *password);
        gboolean (*get_seen) (void *state, const char *id, time_t *last_login, time_t *last_logout,
                              int flags);
	gboolean (*get_messages) (void *state, const Client *client, GList **messages);
	gboolean (*add_message) (void *state, const char *id, const char *message);
        gboolean (*get_buddies) (void *state, Client *client);
        gboolean (*update_buddy) (void *state, const Client *client, const Buddy *buddy);
        gboolean (*add_buddy) (void *state, const Client *client, const Buddy *buddy);
        gboolean (*delete_buddy) (void *state, const Client *client, const Buddy *buddy);
        void *empty;
} DBModule;

void db_set_module (const DBModule *mod);
void *db_init (const GData *params);
void db_shutdown (void *state);
gboolean db_add_client (void *state, const Client *client);
gboolean db_update_login (void *state, const Client *client);
gboolean db_update_logout (void *state, const Client *client);
gboolean db_update_client (void *state, const Client *client);
gboolean db_lookup_client (void *state, const char *id, Client *client);
gboolean db_check_password (void *state, const char *id, const char *password, int flags);
gboolean db_set_password (void *state, const Client *client, const char *password);
gboolean db_get_seen (void *state, const char *id, time_t *last_login, time_t *last_logout,
		      int flags);
gboolean db_get_messages (void *state, const Client *client, GList **messages);
gboolean db_add_message (void *state, const char *id, const char *message);
void db_free_messages (GList *messages);
gboolean db_get_buddies (void *state, Client *client);
gboolean db_update_buddy (void *state, const Client *client, const Buddy *buddy);
gboolean db_add_buddy (void *state, const Client *client, const Buddy *buddy);
gboolean db_delete_buddy (void *state, const Client *client, const Buddy *buddy);

#endif	/* _CLIENT_DB_H_ */
