/* $Id: jenova.c,v 1.70 2001/11/04 01:07:03 robey Exp $
 *
 * Jenova is the say2 central user/channel database server.
 *
 * Copyright (C) 2000, 2001  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 *
 * FIXME:
 * * no timeout on logging in
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <popt.h>
#include <say2/say2.h>
#include "client.h"
#include "say.h"
#include "memory.h"
#include "buddy.h"
#include "client-db.h"
#include "client-sql.h"
#include "client-gdbm.h"

#ifdef DEBUG
#define BANNER	PACKAGE "/" VERSION " (build " JENOVA_BUILD ")"
#else
#define BANNER	PACKAGE "/" VERSION
#endif

/* current say2 protocol version */
#define SAY2_PROTOCOL_MAJOR	0
/* current say2 protocol revision */
#define SAY2_PROTOCOL_MINOR     2
/* protocol revisions within the same major version number should be more
 * or less compatible (a client that speaks protocol 0.3 should work okay
 * against a server that speaks protocol 0.5).  a change in the major version
 * number denotes a change that breaks compatibility.
 */

/* maximum lengths of client-specified strings */
#define MAX_EMAIL	40
#define MAX_STATUS	60
#define MAX_TAGLINE	60
#define MAX_VERSION	60
#define MAX_NICK_LENGTH	32
#define MAX_PASSWORD	32
#define MAX_ADDRESS	60
#define MAX_ICON	4096
#define MAX_ATTRIBUTES	1024

/* maximum answers to a long question */
#define MAX_RESPONSES	50

#define DEFAULT_RCFILE		"/etc/jenovarc"

#define OK_NICK_START(c) ((((c) >= 'A') && ((c) <= 'Z')) || \
                          (((c) >= 'a') && ((c) <= 'z')) || \
                          ((c) == '_') || ((c) == '-'))
#define OK_NICK(c) (OK_NICK_START(c) || \
                    (((c) >= '0') && ((c) <= '9')))

static char *jenova_hostname = NULL;
static int jenova_port = DEFAULT_PORT;
static char *jenova_logfile = NULL;
static int jenova_test = 0;
static int jenova_no_daemon = 0;
static char *db_host = "localhost";
static char *db_user = "robey";
static char *db_pass = "";
static char *db_db = "say2";
static char *jenova_rcfile = DEFAULT_RCFILE;
static char *jenova_certfile = NULL;
static char *jenova_keyfile = NULL;
static char *jenova_motd = "/etc/jenova.motd";
static char *jenova_run_user = "nobody";
static GList *jenova_alias = NULL;
static char *jenova_gdbm_file = NULL;

/* manipulating the user db */
static char *db_lookup_user = NULL;
static char *db_add_user = NULL;
static char *db_add_email = NULL;
static char *db_add_password = NULL;
static char *db_set_privs = NULL;

static void *global_sql = NULL;

static time_t uptime = (time_t)0;

static void *dispatch_table = NULL;
static void *login_dispatch_table = NULL;

static const struct poptOption popt_options[] = {
        { "version", 'v', POPT_ARG_NONE, NULL, 'v', "show version", NULL },
        { "port", 'p', POPT_ARG_INT, &jenova_port, 0, "use alternate say2 port", "port" },
        { "logfile", 'L', POPT_ARG_STRING, &jenova_logfile, 0,
          "use logfile instead of logging to stderr", "filename" },
        { "db-host", '\0', POPT_ARG_STRING, &db_host, 0, "specify SQL server", "hostname" },
        { "db-user", '\0', POPT_ARG_STRING, &db_user, 0, "specify SQL user", "username" },
        { "db-pass", '\0', POPT_ARG_STRING, &db_pass, 0, "specify SQL password", "password" },
        { "db", '\0', POPT_ARG_STRING, &db_db, 0, "specify SQL database", "db-name" },
        { "use-gdbm", 'g', POPT_ARG_STRING, &jenova_gdbm_file, 0, "use gdbm file, not SQL", "filename" },
        { "test", '\0', POPT_ARG_NONE | POPT_ARGFLAG_DOC_HIDDEN, &jenova_test, 0, "test database", NULL },
        { "rcfile", 'f', POPT_ARG_STRING, &jenova_rcfile, 0, "config file to load (default /etc/jenovarc)", "rcfile" },
        { "no-daemon", 'd', POPT_ARG_NONE, &jenova_no_daemon, 0, "don't daemonize", NULL },
        { "as-user", '\0', POPT_ARG_STRING, &jenova_run_user, 0, "run as specified user (if started as root)",
          "username" },
        { "lookup", '\0', POPT_ARG_STRING | POPT_ARGFLAG_DOC_HIDDEN, &db_lookup_user, 0, "", "" },
        { "add", '\0', POPT_ARG_STRING | POPT_ARGFLAG_DOC_HIDDEN, &db_add_user, 0, "", "" },
        { "email", '\0', POPT_ARG_STRING | POPT_ARGFLAG_DOC_HIDDEN, &db_add_email, 0, "", "" },
        { "password" , '\0', POPT_ARG_STRING | POPT_ARGFLAG_DOC_HIDDEN, &db_add_password, 0, "", "" },
        { "set-privs", '\0', POPT_ARG_STRING | POPT_ARGFLAG_DOC_HIDDEN, &db_set_privs, 0, "", "" },
        { NULL, 'h', POPT_ARG_NONE | POPT_ARGFLAG_DOC_HIDDEN, NULL, 'h', NULL, NULL },
        POPT_AUTOHELP
        { NULL, 0, 0, NULL, 0, NULL, NULL }
};


static void client_read (Socket *sock);

static GMainLoop *l_main_loop = NULL;


static gboolean
rcfile_callback (const char *key, const char *value)
{
        if (g_strcasecmp (key, "db-host") == 0) {
                db_host = g_strdup (value);
                return TRUE;
        }
        if (g_strcasecmp (key, "db-user") == 0) {
                db_user = g_strdup (value);
                return TRUE;
        }
        if (g_strcasecmp (key, "db-pass") == 0) {
                db_pass = g_strdup (value);
                return TRUE;
        }
        if (g_strcasecmp (key, "db-name") == 0) {
                db_db = g_strdup (value);
                return TRUE;
        }
        if (g_strcasecmp (key, "logfile") == 0) {
                jenova_logfile = g_strdup (value);
                return TRUE;
        }
        if (g_strcasecmp (key, "hostname") == 0) {
                jenova_hostname = g_strdup (value);
                return TRUE;
        }
        if (g_strcasecmp (key, "alias") == 0) {
                jenova_alias = g_list_prepend (jenova_alias, g_strdup (value));
                return TRUE;
        }
        if (g_strcasecmp (key, "port") == 0) {
                jenova_port = atoi (value);
                return TRUE;
        }
        if (g_strcasecmp (key, "certfile") == 0) {
                jenova_certfile = g_strdup (value);
                return TRUE;
        }
	if (g_strcasecmp (key, "keyfile") == 0) {
		jenova_keyfile = g_strdup (value);
		return TRUE;
	}
        if (g_strcasecmp (key, "motd") == 0) {
                jenova_motd = g_strdup (value);
                return TRUE;
        }
        if (g_strcasecmp (key, "run-as-user") == 0) {
                jenova_run_user = g_strdup (value);
                return TRUE;
        }
        if (g_strcasecmp (key, "gdbm-file") == 0) {
                jenova_gdbm_file = g_strdup (value);
                return TRUE;
        }

        log ("unknown config key '%s'", key);
        return FALSE;
}

static void
send_error (Socket *sock, const char *command, const char *msg)
{
        send_enaml_buildf(sock, "error", "command:%s text:%s", command, msg);
}

/* someone once pointed out that it would make more sense to send the motd
 * as a list of strings, instead of N individual responses.
 */
static void
send_motd (Client *client)
{
        FILE *fp;
        char line[256];

        if ((fp = fopen (jenova_motd, "r")) != NULL) {
                while (! feof (fp)) {
                        fgets (line, 255, fp);
                        if (feof (fp)) {
                                break;
                        }
                        line[255] = 0;
                        while ((strlen (line) > 0) && 
                               ((line[strlen (line)-1] == '\n') || (line[strlen (line)-1] == '\r'))) {
                                line[strlen (line)-1] = 0;
                        }
                        client_sendf(client, "motd", "text:%s", line);
                }
                fclose (fp);
        } else {
                client_sendf(client, "motd", "text:%s", "No MOTD.");
        }
}

static void
send_old_messages (Client *client)
{
        GList *messages, *iter;
        char *msg;

        /* check for stored messages */
        if (db_get_messages(global_sql, client, &messages)) {
                for (iter = g_list_first(messages); iter != NULL; iter = g_list_next(iter)) {
                        msg = (char *)iter->data;
                        client_send_text(client, msg);
                }
                db_free_messages(messages);
        }
}

static void
trigger_login (Client *client, Socket *sock)
{
        if (g_strcasecmp (client->id, client->nick) != 0) {
                log ("%s login as %s", client->id, client->nick);
        } else {
                log ("%s login", client->id);
        }
        send_motd(client);
        db_update_login(global_sql, client);

        if (client->flags & CLIENT_USER) {
                client_notify_signon(client);		/* FIXME */
                db_get_buddies(global_sql, client);
                buddy_list_load(client);
                buddy_update_status(client, TRUE);
                buddy_list_refresh_pending(global_sql, client);

        }
}

static void
trigger_logout (Client *client)
{
        log ("%s logout", client->id);
        client_notify_signoff (client);		/* FIXME */
        buddy_update_status(client, FALSE);
        buddy_list_unload(client);
        db_update_logout (global_sql, client);
}

static void
do_motd (Client *client, const EnamlBlock *ignored, Socket *sock)
{
        send_motd (client);
}

static void
do_server_cert (Client *client, const EnamlBlock *server_cert, Socket *sock)
{
        const char *buffer;
        int length;
        EnamlList *alias_list;
        GList *iter;

        alias_list = enaml_new_list();
        if (jenova_alias != NULL) {
                for (iter = g_list_first (jenova_alias); iter != NULL; iter = g_list_next (iter)) {
                        enaml_list_add_string (alias_list, (char *)iter->data, 0);
                }
        }
        socket_get_cert (sock, &buffer, &length);
        if (buffer != NULL) {
                client_sendf(client, "server-cert", "host:%s alias:%L cert:%x", jenova_hostname,
                             alias_list, buffer, length);
        } else {
                client_sendf(client, "server-cert", "host:%s alias:%L none", jenova_hostname,
                             alias_list);
        }
}

static void
do_start_tls (Client *client, const EnamlBlock *start_tls, Socket *sock)
{
        const char *buffer;
        int length;

        if (socket_is_secure (sock)) {
                send_error (sock, "start-tls", "This connection is already using SSL!");
                return;
        }

        socket_get_cert (sock, &buffer, &length);
        if (buffer == NULL) {
                send_error (sock, "start-tls", "This server is not prepared to use SSL.");
                return;
        }

        /* start TLS! */
        client_sendf(client, "start-tls", NULL);

        if (! socket_begin_ssl (sock, SOCKET_SSL_SERVER)) {
                log ("%s SSL negotiation failed", client->id);
                return;
        }
        log ("%s start SSL", client->id);
}

static void
do_ping (Client *client, const EnamlBlock *ping, Socket *sock)
{
        client_got_ping(client, sock);
}

static void
do_bite_my_ass (Client *client, const EnamlBlock *bite, Socket *sock)
{
        const char *now;

        if (! bite) {
                client_sendf(client, "bite-my-ass", "text:%s", "Now?");
                return;
        }

        now = enaml_get_string(bite, "now", NULL);
        if (now && (now[0] == 'y' || now[0] == 'Y')) {
                client_sendf(client, "CHOMP", "hard:%s", "yes");
                return;
        }
        send_error(sock, "bite-my-ass", "Maybe later...");
}

static void
do_message (Client *client, const EnamlBlock *message, Socket *sock)
{
        Client *dest_client;
        EnamlBlock *out;
        EnamlBlock *message_copy;
	char *msgstr;

        if (! message) {
                send_error(sock, "message", "Message requires arguments.");
                return;
        }

        dest_client = client_get_target(message, CLIENT_USER, &out);
        if (dest_client == NULL) {
		dest_client = client_get_offline_target(global_sql, message, CLIENT_USER, &out);
                if (dest_client == NULL) {
			client_send(client, out);
			enaml_free(out);
			return;
		}
        }

        message_copy = enaml_copy_block (message);

        out = enaml_new();
        enaml_add_block(out, "message", message_copy, ENAML_FREE);
        enaml_add_string(message_copy, "id", client->id, 0);
        enaml_add_long(message_copy, "time", (long) time (NULL));
        enaml_add_string(message_copy, "nick", client->nick, 0);
        enaml_add_string(message_copy, "to-id", dest_client->id, 0);
        enaml_add_string(message_copy, "to-nick", dest_client->nick, 0);

        if (dest_client->flags & (CLIENT_OFFLINE|CLIENT_NO_MSG)) {
                enaml_add_flag(message_copy, "stored");
		msgstr = enaml_compose(out, FALSE);
		db_add_message(global_sql, dest_client->id, msgstr);
		g_free(msgstr);
        } else {
                enaml_drop_key(message_copy, "stored");
		client_send(dest_client, out);
        }

	if (dest_client != client) {
		/* if sending a message to yourself, you only get one copy */
                /* weird corner case here: if you have msgs turned off and send a msg to yourself,
                 * you will get no response at all.
                 */
		client_send(client, out);
	}
        enaml_free(out);

        client_discard(dest_client);

        /* successful message counts as non-idle activity */
        client->activity = time(NULL);
}

static void
do_wall (Client *client, const EnamlBlock *wall, Socket *sock)
{
        EnamlBlock *out;
        const char *text;

        if (wall == NULL) {
                send_error (sock, "wall", "Wall requires arguments.");
                return;
        }
        text = enaml_get_string (wall, "text", NULL);
        if (text == NULL) {
                send_error (sock, "wall", "Required wall field: text");
                return;
        }

        if (!(client->privs & PRIV_WALL)) {
                send_error (sock, "wall", "You don't have permission to send a wall message.");
                return;
        }

        out = enaml_buildf ("wall", "nick:%s id:%s time:%d text:%s",
                            client->nick, client->id, (long)time (NULL), text);
        client_broadcast (out);
        enaml_free (out);
}

static void
send_whois (Client *client, const Client *dest_client, gboolean add_icon)
{
        unsigned long idle, status_time;
        time_t last_login, last_logout;
        char *type;
        EnamlBlock *out, *sub;

        idle = (unsigned long) time (NULL) - (unsigned long) dest_client->activity;
	status_time = (unsigned long) time (NULL) - (unsigned long) dest_client->status_changed;
        type = (dest_client->flags & CLIENT_CHANNEL) ? "channel" :
                ((dest_client->flags & CLIENT_USER) ? "user" : "?");
        if ((long)dest_client->login > 0) {
                /* online */
                if (dest_client->flags & CLIENT_USER) {
                        out = enaml_buildf ("client-info",
                                            "id:%s nick:%s type:%s email:%s login:%d idle:%d "
                                            "tagline:%s mode:%s status:%s status-time:%d "
                                            "client:%s address:%s guest:%F",
                                            dest_client->id, dest_client->nick, type,
                                            dest_client->email, (long)dest_client->login,
                                            (long)idle, dest_client->tagline,
                                            client_mode_string (dest_client), dest_client->status,
                                            (long)status_time,
                                            client_get_version(dest_client), dest_client->address,
					    (dest_client->flags & CLIENT_GUEST) ? TRUE : FALSE);
                } else {	/* channel */
                        out = enaml_buildf ("client-info",
                                            "id:%s type:%s email:%s login:%d idle:%d tagline:%s "
                                            "client:%s address:%s",
                                            dest_client->id, type, dest_client->email,
                                            (long)dest_client->login, (long)idle,
                                            dest_client->tagline, client_get_version(dest_client),
                                            dest_client->address);
                }
        } else {
                /* offline */
                if (! db_get_seen (global_sql, dest_client->id, &last_login, &last_logout,
                                   dest_client->flags)) {
                        last_login = last_logout = (time_t)0L;
                }
                out = enaml_buildf ("client-info",
                                    "offline id:%s type:%s email:%s tagline:%s last-login:%d "
                                    "last-logout:%d",
                                    dest_client->id, type, dest_client->email,
                                    dest_client->tagline, (long)last_login, (long)last_logout);
        }

        if (add_icon && (dest_client->iconsize > 0)) {
                sub = enaml_get_block (out, "client-info");
                g_assert (sub != NULL);
                enaml_add_blob (sub, "icon", dest_client->icon, dest_client->iconsize, 0);
        }

        client_send(client, out);
        enaml_free (out);
}

static void
do_whois (Client *client, const EnamlBlock *whois, Socket *sock)
{
        Client *dest_client;
        EnamlBlock *out;
        const char *id, *type;
        int flags;
        gboolean add_icon;

        if (! whois) {
                send_error (sock, "whois", "Required whois field: nick or id");
                return;
        }

        type = enaml_get_string (whois, "type", "user");
        flags = client_type_to_flags (type);
        if (flags == 0) {
                send_error (sock, "whois", "Type must be user or channel");
                return;
        }
        add_icon = (enaml_type (whois, "icon") == ENAML_KEY_UNDEF) ? FALSE : TRUE;

        dest_client = client_get_target (whois, flags, &out);
        if (dest_client != NULL) {
                send_whois (client, dest_client, add_icon);
                return;
        }

        /* they may want info on an offline user */
        id = enaml_get_string (whois, "id", NULL);
        if (id == NULL) {
                /* must've been looking up by nick: online users only */
                client_send(client, out);
                enaml_free (out);
                return;
        }
        enaml_free (out);
        dest_client = client_lookup_offline (global_sql, id, flags);
        if (dest_client == NULL) {
                /* no such client */
                client_sendf(client, "no-such-client", "type:%s id:%s", type, id);
                return;
        }

        send_whois (client, dest_client, add_icon);
        client_free (dest_client);
        return;
}

static void
do_list_clients_verbose (Client *client, const char *grep, int flags)
{
        GList *list, *iter;
        int count, max, i;
        Client *dest_client;

        list = client_grep (grep);
        max = count = g_list_length (list);
        if (count > MAX_RESPONSES) {
                count = MAX_RESPONSES;
        }
        client_sendf(client, "begin-response", "count:%d total:%d response:%s", count, max, "client-info");
        for (i = 0, iter = g_list_first (list); (i < count) && (iter != NULL); i++, iter = g_list_next (iter)) {
                dest_client = (Client *)(iter->data);
                send_whois (client, dest_client, FALSE);
        }
        client_sendf(client, "end-response", "response:%s", "client-info");
}

static void
do_list_clients (Client *client, const EnamlBlock *list_clients, Socket *sock)
{
        EnamlBlock *out, *members;
        const char *type, *grep = NULL;
        gboolean use_ids = FALSE;
        int flags;

        if (list_clients == NULL) {
                send_error (sock, "list-clients", "Required list-clients field: type");
                return;
        }

        type = enaml_get_string (list_clients, "type", NULL);
        if (type == NULL) {
                send_error (sock, "list-clients", "Required list-clients field: type");
                return;
        }
        flags = client_type_to_flags (type);
        if ((flags == 0) && (strcmp (type, "all") != 0)) {
                send_error (sock, "list-clients",
                            "list-clients field 'type' must be one of (user, channel, all)");
                return;
        }

        grep = enaml_get_string (list_clients, "grep", NULL);
        if (enaml_type (list_clients, "verbose") == ENAML_KEY_FLAG) {
                do_list_clients_verbose (client, grep, flags);
                return;
        }
        if (enaml_type (list_clients, "by-id") == ENAML_KEY_FLAG) {
                use_ids = TRUE;
        }

        out = enaml_new ();
        members = enaml_new ();
        enaml_add_block (out, "members", members, ENAML_FREE);
        enaml_add_string (members, "type", type, 0);
        if (use_ids) {
                client_get_id_list (members, "list", grep, flags);
        } else {
                client_get_list (members, "list", grep, flags);
        }

        client_send(client, out);
        enaml_free (out);
}

static void
do_logout (Client *client, const EnamlBlock *logout, Socket *sock)
{
        send_enaml_buildf(sock, "logout", "text:%s", "Goodbye!");
        socket_close(sock);
}

static void
do_status (Client *client, const EnamlBlock *status, Socket *sock)
{
        const char *text, *mode_text, *msgs_text;
        char *old_text;
        EnamlBlock *block;
        ClientMode old_mode;
        int old_msgs;

        if (status == NULL) {
                block = client_make_status(client, TRUE);
                client_send(client, block);
                enaml_free(block);
                return;
        }

        old_text = client->status ? g_strdup (client->status) : NULL;
        text = enaml_get_string (status, "text", NULL);
        if (text) {
                g_free (client->status);
                client->status = strlen (text) ? g_strndup (text, MAX_STATUS) : NULL;
        }

        old_mode = client->mode;
        mode_text = enaml_get_string (status, "mode", NULL);
        if (mode_text) {
                if (strcasecmp (mode_text, "active") == 0) {
                        client->mode = CLIENT_MODE_ACTIVE;
                }
                if (strcasecmp (mode_text, "busy") == 0) {
                        client->mode = CLIENT_MODE_BUSY;
                }
                if (strcasecmp (mode_text, "away") == 0) {
                        client->mode = CLIENT_MODE_AWAY;
                }
        }

        old_msgs = (client->flags & CLIENT_NO_MSG);
        msgs_text = enaml_get_string(status, "messages", NULL);
        if (msgs_text != NULL) {
                if (strcasecmp(msgs_text, "on") == 0) {
                        client->flags &= ~CLIENT_NO_MSG;
                } else {
                        client->flags |= CLIENT_NO_MSG;
                }
        }

        if ((old_mode != client->mode) ||
            (old_msgs != (client->flags & CLIENT_NO_MSG)) ||
            (old_text && !client->status) || (!old_text && client->status) ||
            (old_text && client->status && (strcmp (old_text, client->status) != 0))) {
                /* something changed -- broadcast the new status */
	        /* btw, a change in status preemptively counts as activity */
	        client->activity = time (NULL);
                if (1) {
                        /* FIXME */
                        block = client_make_status(client, TRUE);
                        client_broadcast (block);
                        enaml_free (block);
                }
                buddy_update_status(client, TRUE);
		client->status_changed = time (NULL);
        }

        if (!(client->flags & CLIENT_NO_MSG) && old_msgs) {
                /* deliver any pending messages */
                send_old_messages(client);
        }

        g_free (old_text);
}

/* if it returns false, an error has already been sent */
static gboolean
ok_nick (Client *client, const char *id, const char *nick)
{
        int bad = 0;
        const char *p;

        if (client_by_nick (nick)) {
                client_sendf(client, "bad-nick", "nick:%s text:%s",
                             nick, "Nickname is already in use.");
                return FALSE;
        }
        if ((id != NULL) && (strcasecmp (id, nick) == 0)) {
                /* always allowed to use your own id */
                return TRUE;
        }
        if (db_lookup_client (global_sql, nick, NULL)) {
                client_sendf(client, "bad-nick", "nick:%s text:%s",
                             nick, "Nickname belongs to a registered user.");
                return FALSE;
        }
        if (strlen (nick) > MAX_NICK_LENGTH) {
                client_sendf(client, "bad-nick", "nick:%s text:%s",
                             nick, "Nickname is too long.");
                return FALSE;
        }

        bad = 0;
        if (! OK_NICK_START(*nick)) {
                bad = 1;
        }
        for (p = nick+1; *p && !bad; p++) {
                if (! OK_NICK(*p)) {
                        bad = 1;
                }
        }
        if (bad) {
                client_sendf(client, "bad-nick", "nick:%s text:%s",
                             nick, "Nickname contains an illegal symbol.");
                return FALSE;
        }

        return TRUE;
}

static char *
get_clean_email_tagline (const char *in, int limit)
{
        const char *p;
        gboolean okay = TRUE;

        for (p = in; *p && okay; p++) {
                if ((*p < ' ') || (*p >= 127)) {
                        okay = FALSE;
                }
        }
        if (! okay) {
                return NULL;
        }

        return g_strndup (in, limit);
}

static gboolean
address_is_okay (const char *address)
{
        const char *p;
        char *host;
        unsigned long port;

        if (strlen (address) > MAX_ADDRESS) {
                return FALSE;
        }
        p = strchr (address, ':');
        if ((p == NULL) || (p == address)) {
                return FALSE;
        }
        port = strtoul (p+1, &host, 10);
        if (*host != '\0') {
                return FALSE;
        }
        if ((port == 0) || (port > 65535)) {
                return FALSE;
        }
        host = g_strndup (address, p-address);
        if (! socket_dns_resolves (host)) {
                g_free (host);
                return FALSE;
        }
        g_free (host);
        return TRUE;
}

static void
do_login (Client *client, const EnamlBlock *login, Socket *sock)
{
        const char *nick, *alt_nick, *email, *tagline, *version, *password, *address;
        Client *old_client;
        gboolean attached = FALSE;
        gboolean replicant = FALSE;
        gboolean send_messages = FALSE;
        int flag = CLIENT_USER;
        EnamlBlock *block;

        if (login == NULL) {
                send_error (sock, "login", "Login requires arguments");
                return;
        }

        if (client->nick != NULL) {
                send_error(sock, "login", "You are already logged in.");
                return;
        }

	alt_nick = NULL;
        nick = enaml_get_string (login, "channel", NULL);
        if (nick != NULL) {
                flag = CLIENT_CHANNEL;
        } else {
                nick = enaml_get_string (login, "id", NULL);
                if (! nick) {
                        nick = enaml_get_string (login, "nick", NULL);
                        if (! nick) {
                                send_error (sock, "login", "Required login field: nick or channel");
                                return;
                        }
                } else {
                        /* registered id may want to use a nick */
                        alt_nick = enaml_get_string(login, "nick", NULL);
                }
        }

        version = enaml_get_string (login, "client", NULL);

        address = enaml_get_string (login, "address", NULL);
        if ((address == NULL) && (flag == CLIENT_CHANNEL)) {
                send_error (sock, "login", "Required login field for channels: address");
                return;
        }
        if (address != NULL) {
                if (! address_is_okay (address)) {
                        client_sendf(client, "bad-address", "address", address);
                        return;
                }
        }

	/* you can't attach to a session until it's logged in, so we can be sure that
	 * there's only one socket on this Client right now.
	 */

        password = enaml_get_string (login, "password", NULL);
        if (password != NULL) {
                if (! db_check_password (global_sql, nick, password, flag)) {
                        send_error (sock, "login", "Bad password.");
                        return;
                }
                old_client = client_by_id (nick, flag);
                if (old_client != NULL) {
                        if (flag == CLIENT_CHANNEL) {
                                send_error (sock, "login", "You are already logged in.");
                                return;
                        }

                        if (enaml_type(login, "replicant") != ENAML_KEY_UNDEF) {
                                /* replicant */
                                replicant = TRUE;
                                log("%s reattached to %s (replicant)", client->id, old_client->id);
                        } else {
                                /* successful reconnect, but the client was on before: detach previous session */
                                client_sendf(old_client, "logout", "text:%s detached:%s",
                                             "This session has been detached.",
					     client_host_by_socket(client, sock));
                                log ("%s reattached to %s", client->id, old_client->id);
                        }
			if (! client_move_socket(old_client, client, sock)) {
				send_error (sock, "login", "Too many replicants.");
				return;
			}
			if (! replicant) {
                                client_close_all_sockets_but(old_client, sock);
			}

                        /* attach to the old client */
                        client_free (client);
                        client = old_client;
                        socket_set_data (sock, client);
                        client->activity = time (NULL);
                        attached = TRUE;
                        if (client->flags & CLIENT_NO_MSG) {
                                /* start receiving messages again */
                                send_messages = TRUE;
                                client->flags &= ~CLIENT_NO_MSG;
                        }
                } else {
                        /* good password!  ...shortcut to filling in data, etc */
                        client->flags &= ~(CLIENT_USER|CLIENT_CHANNEL);
                        client->flags |= flag;
                        db_lookup_client (global_sql, nick, client);
                        if (alt_nick != NULL) {
				if ((g_strcasecmp(nick, alt_nick) != 0) &&
				    !ok_nick(client, client->id, alt_nick)) {
                                        return;
                                }
                                g_free(client->nick);
                                client->nick = g_strdup(alt_nick);
                        }
                }
        } else if (flag == CLIENT_CHANNEL) {
                send_error (sock, "login", "Required login field: password");
                return;
        } else {
                /* login without actually having an account: this should go away (someday...) */
                email = enaml_get_string (login, "email", NULL);
                if (! email) {
                        send_error (sock, "login", "Required login field: email");
                        return;
                }

                if (! ok_nick (client, NULL, nick)) {
                        return;
                }

                tagline = enaml_get_string (login, "tagline", NULL);

                client->nick = g_strdup (nick);
                client->email = get_clean_email_tagline (email, MAX_EMAIL);
                if (tagline) {
                        client->tagline = get_clean_email_tagline (tagline, MAX_TAGLINE);
                }
                client->flags |= CLIENT_GUEST;
        }

        if (version != NULL) {
                client_set_version(client, get_clean_email_tagline (version, MAX_VERSION));
        }
        if (address != NULL) {
		g_free(client->address);
                client->address = g_strdup (address);
        }

        if (flag == CLIENT_CHANNEL) {
                client_sendf(client, "welcome", "channel:%s", client->id);
        } else {
                client_sendf(client, "welcome", "nick:%s id:%s attached:%F replicant:%F host:%s",
                             client->nick, client->id, attached, replicant,
			     client_host_by_socket(client, sock));
        }
        socket_set_read_fn (sock, client_read);

        /* send status update (in case this is an existing session) */
        block = client_make_status(client, TRUE);
        client_send(client, block);
        enaml_free(block);

        if (! attached) {
                trigger_login (client, sock);
        } else if (send_messages) {
                buddy_update_status(client, TRUE);
                send_old_messages(client);
        }
}

static void
do_boot_user (Client *client, const EnamlBlock *boot, Socket *sock)
{
	const char *text, *type;
	Client *target;
	EnamlBlock *out;
        int flags;

	if (boot == NULL) {
		send_error (sock, "boot-user", "boot-user requires an id or nick");
		return;
	}
        type = enaml_get_string (boot, "type", "user");
        flags = client_type_to_flags (type);
        if (flags == 0) {
                send_error (sock, "boot-user", "Type must be user or channel");
                return;
        }
        target = client_get_target (boot, flags, &out);
        if (! target) {
                client_send(client, out);
                enaml_free (out);
                return;
        }

        if (!(client->privs & PRIV_BOOTUSER)) {
                send_error (sock, "boot-user", "You don't have permission to boot users.");
                return;
        }

	text = enaml_get_string (boot, "text", NULL);
	out = enaml_buildf ("boot-user", "nick:%s id:%s target-nick:%s target-id:%s text:%s",
			    client->nick, client->id, target->nick, target->id, text);
	client_broadcast (out);

        client_close_all_sockets(target);
}

static void
do_add_client (Client *client, const EnamlBlock *add, Socket *sock)
{
        const char *id, *type, *email, *password;
        Client *new_client;
        int flags;

        if (add == NULL) {
                send_error (sock, "add-client",
                            "Required add-client fields: type, id, email, and password");
                return;
        }

        id = enaml_get_string (add, "id", NULL);
        type = enaml_get_string (add, "type", NULL);
        email = enaml_get_string (add, "email", NULL);
        password = enaml_get_string (add, "password", NULL);
        if ((id == NULL) || (type == NULL) || (email == NULL) || (password == NULL)) {
                send_error (sock, "add-client",
                            "Required add-client fields: type, id, email, and password");
                return;
        }
        flags = client_type_to_flags (type);
        if (flags == 0) {
                send_error (sock, "add-client", "Type must be user or channel");
                return;
        }

        if (!(client->privs & PRIV_ADDUSER)) {
                send_error (sock, "add-client", "You don't have permission to add clients");
                return;
        }

        if (! ok_nick(client, NULL, id)) {
                return;
        }

        new_client = client_new(FALSE);
        new_client->id = g_strdup (id);
        new_client->email = g_strdup (email);
        new_client->flags = flags;

        if (! db_add_client (global_sql, new_client)) {
                send_error (sock, "add-client", "Adding that client failed, for some reason.");
                return;
        }

        if (! db_set_password (global_sql, new_client, password)) {
                send_error (sock, "add-client",
                            "Added the client, but couldn't set the password.");
                return;
        }

        client_free (new_client);
        client_sendf(client, "add-client", "ok");
}

static void
do_prefs (Client *client, const EnamlBlock *prefs, Socket *sock)
{
        const char *tagline, *notify, *password;
        char *c_tagline;
        const char *icon;
        int iconsize;
        gboolean have_icon;

        if (client->flags & CLIENT_GUEST) {
                send_enaml_buildf(sock, "no-user-record", NULL);
                return;
        }

        if (prefs == NULL) {
                tagline = notify = password = NULL;
                have_icon = FALSE;
        } else {
                tagline = enaml_get_string (prefs, "tagline", NULL);
                notify = enaml_get_string (prefs, "notify", NULL);
                password = enaml_get_string (prefs, "password", NULL);
                have_icon = (enaml_type (prefs, "icon") == ENAML_KEY_VALUE);
        }

        if (tagline != NULL) {
                c_tagline = get_clean_email_tagline (tagline, MAX_TAGLINE);
                if (c_tagline != NULL) {
                        g_free (client->tagline);
                        if (strlen (c_tagline) > 0) {
                                client->tagline = c_tagline;
                        } else {
                                g_free (c_tagline);
                                client->tagline = NULL;
                        }
                }
        }

        if (notify != NULL) {
                if (g_strcasecmp (notify, "on") == 0) {
                        client->flags |= CLIENT_NOTIFY;
                } else {
                        client->flags &= ~CLIENT_NOTIFY;
                }
        }

        if (password != NULL) {
                if (strlen (password) > MAX_PASSWORD) {
                        send_error (sock, "prefs", "Password is too long.");
                        return;
                }
                if (! db_set_password (global_sql, client, password)) {
                        send_error (sock, "prefs", "Unable to set your password.");
                        return;
                }
        }

        if (have_icon) {
                iconsize = enaml_get_blob (prefs, "icon", &icon);
                if (iconsize > MAX_ICON) {
                        send_error (sock, "prefs", "Icon is too long.");
                        return;
                }
                if (client->icon != NULL) {
                        g_free (client->icon);
                        client->icon = NULL;
                        client->iconsize = 0;
                }
                if (iconsize > 0) {
                        client->iconsize = iconsize;
                        client->icon = g_malloc (iconsize);
                        memcpy (client->icon, icon, iconsize);
                }
        }

        if ((tagline != NULL) || (notify != NULL) || have_icon) {
                if (! db_update_client (global_sql, client)) {
                        send_error (sock, "prefs", "Couldn't update your record right now.");
                        return;
                }
        }

        client_sendf(client, "prefs",
                     "tagline:%s notify:%s password-changed:%F icon:%F",
                     client->tagline, (client->flags & CLIENT_NOTIFY) ? "on" : "off",
                     password != NULL, client->icon != NULL);
}

static void
do_change_nick (Client *client, const EnamlBlock *change, Socket *sock)
{
        const char *nick;
        EnamlBlock *block;

        if (change == NULL) {
                send_error (sock, "change-nick", "Required change-nick field: nick");
                return;
        }

        nick = enaml_get_string (change, "nick", NULL);
        if (! nick) {
                send_error (sock, "change-nick", "Required change-nick field: nick");
                return;
        }

        if (g_strcasecmp (nick, client->nick) == 0) {
                client_sendf(client, "bad-nick", "nick:%s text:%s",
                             nick, "That's your current nick.");
                return;
        }
        if (! ok_nick(client, client->id, nick)) {
                return;
        }

        block = enaml_buildf ("change-nick", "id:%s old-nick:%s nick:%s",
                              client->id, client->nick, nick);
        client_broadcast (block);
        enaml_free (block);

        g_free (client->nick);
        client->nick = g_strdup (nick);
        client->activity = time (NULL);
}

static void
do_set_attributes (Client *client, const EnamlBlock *set_attr, Socket *sock)
{
        const char *key;
        EnamlIter iter;
	EnamlList *list;
        EnamlBlock *block, *out;
        char *old_attributes, *new_attributes;

        if (client->flags & CLIENT_GUEST) {
                send_enaml_buildf(sock, "no-user-record", NULL);
                return;
        }

        if (set_attr == NULL) {
                /* do nothing. */
                client_sendf(client, "set-attributes", NULL);
                return;
        }

        /* also make a list of changed keys */
        list = enaml_new_list ();
        old_attributes = enaml_compose (client->attributes, FALSE);

        ENAML_FOREACH (key, set_attr, iter) {
                /* remove any old key */
                enaml_drop_key (client->attributes, key);
                switch (enaml_type (set_attr, key)) {
                case ENAML_KEY_UNDEF:
                        g_assert_not_reached ();
                case ENAML_KEY_FLAG:
                        /* (kinda awkward) way to remove a key */
                        break;
                default:
                        if (client->attributes == NULL) {
                                client->attributes = enaml_new ();
                        }
                        enaml_copy_key (set_attr, client->attributes, key);
                        enaml_list_add_string (list, key, ENAML_COPY);
                        break;
                }
        }

        /* check that we didn't run out of space */
        new_attributes = enaml_compose (client->attributes, FALSE);
        if (strlen (new_attributes) > MAX_ATTRIBUTES) {
                g_free (new_attributes);
                enaml_free (client->attributes);
                client->attributes = enaml_parse (old_attributes);
                g_free (old_attributes);
                send_error (sock, "set-attributes", "Attribute block would be too large.");
                return;
        }

        g_free (old_attributes);
        g_free (new_attributes);

        if ((client->attributes != NULL) && (enaml_block_size (client->attributes) == 0)) {
                /* just kill it off */
                enaml_free (client->attributes);
                client->attributes = NULL;
        }

        if (! db_update_client (global_sql, client)) {
                send_error (sock, "set-attributes", "Couldn't update your record right now.");
                return;
        }

        /* response: list of keys we added/changed (but not deleted -- should those be included too?) */
        block = enaml_new ();
        enaml_add_list (block, "attributes", list, ENAML_FREE);
        out = enaml_new ();
        enaml_add_block (out, "set-attributes", block, ENAML_FREE);
        client_send(client, out);
        enaml_free (out);
}

static void
do_get_attributes (Client *client, const EnamlBlock *get_attr, Socket *sock)
{
        const char *id, *key, *type;
        int i;
        int flags;
        Client *dest_client;
        EnamlBlock *response;
        EnamlList *list;

        if (get_attr == NULL) {
                send_error (sock, "get-attributes", "Required get-attributes field: id");
                return;
        }

        id = enaml_get_string (get_attr, "id", NULL);
        if (id == NULL) {
                send_error (sock, "get-attributes", "Required get-attributes field: id");
                return;
        }

        type = enaml_get_string (get_attr, "type", "user");
        flags = client_type_to_flags (type);
        if (flags == 0) {
                send_error (sock, "get-attributes", "Type must be user or channel");
                return;
        }

        dest_client = client_get_target (get_attr, flags, &response);
        if (dest_client == NULL) {
		dest_client = client_get_offline_target(global_sql, get_attr, flags, &response);
                if (dest_client == NULL) {
			client_send(client, response);
			enaml_free(response);
                        return;
                }
        }

        list = enaml_get_list (get_attr, "attribute-list");
        if (list == NULL) {
                client_sendf(client, "client-attributes", "type:%s id:%s attributes:%!B",
                             type, id, dest_client->attributes);
        } else {
                /* list of specific attrs to show... */
                response = enaml_new ();
                for (i = 0; i < enaml_list_length (list); i++) {
                        key = enaml_list_get_string (list, i, NULL);
                        if (key == NULL) {
                                continue;
                        }
                        enaml_copy_key (dest_client->attributes, response, key);
                }

                client_sendf(client, "client-attributes", "type:%s id:%s attributes:%!B",
                             type, id, response);
                enaml_free (response);
        }

        client_discard(dest_client);
}

static void
do_subscribe (Client *client, const EnamlBlock *subscribe, Socket *sock)
{
        Client *dest_client;
        EnamlBlock *out;
        const char *text;

	if (subscribe == NULL) {
                send_error(sock, "subscribe", "Required subscribe field: id");
                return;
        }

        dest_client = client_get_target(subscribe, CLIENT_USER, &out);
        if (dest_client == NULL) {
		dest_client = client_get_offline_target(global_sql, subscribe, CLIENT_USER, &out);
                if (dest_client == NULL) {
                        client_send(client, out);
                        enaml_free(out);
                        return;
                }
        }

        text = enaml_get_string(subscribe, "text", NULL);
        buddy_list_handle_subscribe(global_sql, client, sock, dest_client->id, text);
        client_discard(dest_client);
}

static void
do_unsubscribe (Client *client, const EnamlBlock *subscribe, Socket *sock)
{
        Client *dest_client;
        EnamlBlock *out;

        if (subscribe == NULL) {
                send_error(sock, "unsubscribe", "Required unsubscribe field: id");
                return;
        }

        dest_client = client_get_target(subscribe, CLIENT_USER, &out);
        if (dest_client == NULL) {
                dest_client = client_get_offline_target(global_sql, subscribe, CLIENT_USER, &out);
                if (dest_client == NULL) {
                        client_send(client, out);
                        enaml_free(out);
                        return;
                }
        }

        buddy_list_handle_unsubscribe(global_sql, client, sock, dest_client->id);
        client_discard(dest_client);
}

static void
do_subscription_accepted (Client *client, const EnamlBlock *subscribe, Socket *sock)
{
        Client *dest_client;
	EnamlBlock *out;
        const char *text;

        if (subscribe == NULL) {
                send_error(sock, "subscription-accepted", "Required subscription-accepted field: id");
                return;
        }

        dest_client = client_get_target(subscribe, CLIENT_USER, &out);
        if (dest_client == NULL) {
                dest_client = client_get_offline_target(global_sql, subscribe, CLIENT_USER, &out);
                if (dest_client == NULL) {
                        client_send(client, out);
                        enaml_free(out);
                        return;
                }
        }

        text = enaml_get_string(subscribe, "text", NULL);
        buddy_list_handle_accept(global_sql, client, sock, dest_client->id, text);
        client_discard(dest_client);
}

static void
do_subscription_denied (Client *client, const EnamlBlock *subscribe, Socket *sock)
{
        Client *dest_client;
	EnamlBlock *out;
        const char *text;

        if (subscribe == NULL) {
                send_error(sock, "subscription-denied", "Required subscription-denied field: id");
                return;
        }

        dest_client = client_get_target(subscribe, CLIENT_USER, &out);
        if (dest_client == NULL) {
                dest_client = client_get_offline_target(global_sql, subscribe, CLIENT_USER, &out);
                if (dest_client == NULL) {
                        client_send(client, out);
                        enaml_free(out);
                        return;
                }
        }

        text = enaml_get_string(subscribe, "text", NULL);
        buddy_list_handle_denied(global_sql, client, sock, dest_client->id, text);
        client_discard(dest_client);
}

static void
do_list_subscriptions (Client *client, const EnamlBlock *subscribe, Socket *sock)
{
        buddy_list_refresh(client);
        buddy_list_refresh_reverse(client);
}

static void
do_refresh_status (Client *client, const EnamlBlock *refresh, Socket *sock)
{
        client_show_buddy_status(client);
}

static void
do_server_info (Client *client, const EnamlBlock *info, Socket *sock)
{
        char *text;
        int clients;

        clients = client_count ();

        text = g_strdup_printf ("%d client%s connected",
                                clients, (clients == 1 ? "" : "s"));
        client_sendf(client, "server-info", "text:%s server:%s version:%s uptime:%d",
                     text, PACKAGE, VERSION, (long)uptime);
        g_free (text);
}

static void
send_debug (Client *client, const char *message)
{
        client_sendf(client, "message", "id:%s nick:%s to-id:%s to-nick:%s text:%s",
                     "(debug)", "(debug)", client->id, client->nick, message);
}

static void
do_debug (Client *client, const EnamlBlock *debug, Socket *sock)
{
        GList *list, *iter;
        Client *c;
        char *str;

        if (!(client->privs & PRIV_DEBUG)) {
                send_error (sock, "debug", "You don't have permission to debug the server.");
                return;
        }

        if (debug == NULL) {
                send_error (sock, "debug", "Debug: client-list sockets");
                return;
        }

        if (enaml_type (debug, "client-list") != ENAML_KEY_UNDEF) {
                list = client_grep (NULL);
                send_debug (client, "Online clients:");
                for (iter = g_list_first (list); iter != NULL; iter = g_list_next (iter)) {
                        c = (Client *)(iter->data);
                        str = client_to_string (c, TRUE);
                        send_debug (client, str);
                        g_free (str);
                }
                send_debug (client, "--- End of list.");
                return;
        }

        if (enaml_type (debug, "sockets") != ENAML_KEY_UNDEF) {
                send_debug (client, "Socket debug dump:");
                list = socket_get_debug ();
                for (iter = g_list_first (list); iter != NULL; iter = g_list_next (iter)) {
                        send_debug (client, (char *)iter->data);
                }
                socket_free_debug (list);
                send_debug (client, "--- End of dump.");
                return;
        }

        if (enaml_type(debug, "buddy-table") != ENAML_KEY_UNDEF) {
                send_debug(client, "Buddy table dump:");
                list = buddy_table_dump();
                for (iter = g_list_first(list); iter != NULL; iter = g_list_next(iter)) {
                        send_debug(client, (char *)iter->data);
                }
                buddy_table_dump_free(list);
                send_debug(client, "--- End of dump.");
                return;
        }

        send_error (sock, "debug", "Debug: client-list sockets buddy-table");
}


static DispatchItem command_table[] = {
        { "motd", &do_motd },
        { "server-cert", &do_server_cert },
        { "start-tls", &do_start_tls },
        { "ping", &do_ping },
        { "bite-my-ass", &do_bite_my_ass },
        { "message", &do_message },
        { "wall", &do_wall },
        { "whois", &do_whois },
        { "list-clients", &do_list_clients },
        { "logout", &do_logout },
        { "status", &do_status },
        { "boot-user", &do_boot_user },
        { "add-client", &do_add_client },
        { "prefs", &do_prefs },
        { "change-nick", &do_change_nick },
        { "set-attributes", &do_set_attributes },
        { "get-attributes", &do_get_attributes },
	{ "subscribe", &do_subscribe },
	{ "unsubscribe", &do_unsubscribe },
        { "subscription-accepted", &do_subscription_accepted },
        { "subscription-denied", &do_subscription_denied },
        { "list-subscriptions", &do_list_subscriptions },
        { "refresh-status", &do_refresh_status },
        { "server-info", &do_server_info },
        { "debug", &do_debug },
        /* say.c */
        { "say", &do_say },
        { "history", &do_history },
        /* memory.c */
        { "list-memories", &do_list_memories },
        { "clear-memory", &do_clear_memory },
        { "new-memory", &do_new_memory },
        { "dump-memory", &do_dump_memory },
        { NULL, NULL }
};

static DispatchItem login_command_table[] = {
        { "server-cert", &do_server_cert },
        { "start-tls", &do_start_tls },
        { "ping", &do_ping },
        { "logout", &do_logout },
        { "login", &do_login },
        { NULL, NULL }
};


static void
client_read (Socket *sock)
{
        Client *client = (Client *) socket_get_data (sock);
        EnamlBlock *in;
        const EnamlBlock *sub;
        EnamlIter iter;
        const char *key;
        char *line;

        g_assert (IS_CLIENT (client));

        line = NULL;
        in = NULL;
        do {
                g_free (line);
                if (in != NULL) {
                        enaml_free (in);
                }

                line = socket_getline (sock);
                if (! line) {
                        /* still buffering */
                        return;
                }

                in = enaml_parse (line);
                if (! in) {
			send_enaml_buildf(sock, "error", "text:%s line:%s",
                                          "enaml could not parse that line", line);
                        goto out;
                }

                for (key = enaml_iter_first (in, &iter); key; key = enaml_iter_next (&iter)) {
                        sub = enaml_get_block (in, key);
                        if (! dispatch_go (dispatch_table, key, client, (void *)sub, sock)) {
                                send_enaml_buildf (sock, "error", "text:%s command:%s",
                                                   "Sorry, I don't understand that command.", key);
                                log ("%s unknown: %s", client->id, line);
                                goto out;
                        }
                }
        } while (line != NULL);

out:
        g_free (line);
        if (in != NULL) {
                enaml_free (in);
        }
}


static void
client_login_read (Socket *sock)
{
        Client *client = (Client *) socket_get_data (sock);
        EnamlBlock *in;
        const EnamlBlock *sub;
        EnamlIter iter;
        const char *key;
        char *line;

        g_assert (IS_CLIENT (client));

        line = NULL;
        in = NULL;
        do {
                g_free (line);
                if (in != NULL) {
                        enaml_free (in);
                }

                line = socket_getline (sock);
                if (! line) {
                        /* still buffering */
                        return;
                }

                in = enaml_parse (line);
                if (! in) {
                        send_error (sock, NULL, "enaml could not parse that line :(");
                        goto out;
                }

                for (key = enaml_iter_first (in, &iter); key; key = enaml_iter_next (&iter)) {
                        sub = enaml_get_block (in, key);
                        if (! dispatch_go (login_dispatch_table, key, client, (void *)sub, sock)) {
                                /* only login allowed here */
                                send_error (sock, key, "You need to login first!");
                                log ("%s pre-login: %s", client->id, line);
                        }
                }

                if (client->nick != NULL) {
                        /* client has successfully logged in.
                         * don't read any more of the buffer, cause other commands will be only
                         * be meaningful to client_read().
                         */
                        break;
                }
        } while (line != NULL);

out:
        g_free (line);
        if (in != NULL) {
                enaml_free (in);
        }

        if (client->nick != NULL) {
                /* give the normal command processor a chance to handle anything buffered */
                client_read(sock);
        }
}

static void
client_eof (Socket *sock)
{
        Client *client = (Client *) socket_get_data (sock);
        char *host;

        if (client == NULL) {
                /* previously detached connection */
                return;
        }

        g_assert (IS_CLIENT (client));

        host = g_strdup(client_host_by_socket(client, sock));
        socket_set_data (sock, NULL);
        client_remove_socket(client, sock);

        if (! client_has_sockets(client)) {
                /* all sockets have disconnected -- bye-bye client! */
                if (client->nick) {
                        trigger_logout (client);
                } else {
                        log ("%s lost client", client->id);
                }
                client_free (client);
        } else {
                client_sendf(client, "replicant-terminated", "host:%s", host);
        }

        g_free(host);
}

/* incoming new connection */
static void
new_connect (Socket *sock, void *user_data, const char *host, int port)
{
        Client *client;
        char *text;
        int clients;

        clients = client_count ();

        send_enaml_buildf (sock, "hello", "protocol:%2d", SAY2_PROTOCOL_MAJOR, SAY2_PROTOCOL_MINOR);
        text = g_strdup_printf ("%d client%s connected",
                                clients, (clients == 1 ? "" : "s"));
        send_enaml_buildf (sock, "server-info", "text:%s server:%s version:%s uptime:%d",
                           text, PACKAGE, VERSION, (long)uptime);
        g_free (text);

        client = client_new(TRUE);
        client_new_connection(client, sock, host, port);

        socket_set_read_fn (sock, client_login_read);
        socket_set_eof_fn (sock, client_eof);

        log ("%s new connection from %s:%d", client->id, host, port);
}

static gboolean
idle_check (void *unused)
{
        client_ping_all ();
        return TRUE;	/* keep calling me */
}

static gboolean
read_file_to_buffer (const char *filename, char **buffer, int *length)
{
        int fd;
        struct stat statbuf;

        *buffer = NULL;
        *length = 0;

        if (stat (filename, &statbuf) != 0) {
                return FALSE;
        }

        fd = open (filename, O_RDONLY);
        if (fd < 0) {
                return FALSE;
        }

        *buffer = g_malloc (statbuf.st_size);
        *length = statbuf.st_size;
        if (read (fd, *buffer, *length) < *length) {
                g_free (*buffer);
                *buffer = NULL;
                *length = 0;
                close (fd);
                return FALSE;
        }

        close (fd);
        return TRUE;
}

static void
set_keys_on_socket (Socket *sock)
{
        char *buffer;
        int length;

        if ((jenova_certfile == NULL) && (jenova_keyfile == NULL)) {
                /* no SSL */
                return;
        }

        if ((jenova_certfile == NULL) || (jenova_keyfile == NULL)) {
                if (jenova_certfile == NULL) {
                        log ("no certfile specified");
                } else {
                        log ("no keyfile specified");
                }
                log ("You can't do SSL without both a certificate AND a private key.");
                exit (1);
        }

        if (! read_file_to_buffer (jenova_certfile, &buffer, &length)) {
                log ("BAD: unable to load certfile %s", jenova_certfile);
                exit (1);
        }
        socket_set_cert (sock, buffer, length);
        g_free (buffer);

        if (! read_file_to_buffer (jenova_keyfile, &buffer, &length)) {
                log ("BAD: unable to read private key file %s", jenova_keyfile);
                exit (1);
        }
        socket_set_private_key (sock, buffer, length);
        g_free (buffer);

       	if (! socket_init_ssl (NULL, NULL)) {
                log ("WARNING: no SSL support");
        }
}

static void
term_signal (int unused)
{
        log (PACKAGE "/" VERSION " shutting down.");
        g_main_quit (l_main_loop);
}

int
main (int argc, char **argv)
{
        Socket *listen;
	poptContext popt_context;
        int option;
        GData *params;

	popt_context = poptGetContext("jenova", argc, (const char **)argv, popt_options, 0);
	while ((option = poptGetNextOpt(popt_context)) >= 0) {
		switch (option) {
                case 'h':
                        fprintf (stderr, "\n" BANNER "\n\n");
                        poptPrintUsage (popt_context, stderr, 0);
                        exit (0);
                        break;
                case 'v':
                        fprintf (stderr, "\n" BANNER "\n\n");
                        exit (0);
                        break;
                }
        }
        poptFreeContext (popt_context);

        /* send logs to stderr during the rcfile processing */
        log_set (stderr);

        if (jenova_rcfile && *jenova_rcfile) {
                if (! rcfile_read (jenova_rcfile, rcfile_callback)) {
                        if (strcmp (jenova_rcfile, DEFAULT_RCFILE) == 0) {
                                log ("warning: couldn't read %s -- attempting to hobble along with defaults!",
                                     jenova_rcfile);
                        } else {
                                exit (1);
                        }
                }
        }

        dispatch_table = dispatch_create (command_table);
        login_dispatch_table = dispatch_create (login_command_table);

        if (jenova_hostname == NULL) {
                jenova_hostname = g_malloc (128);
                if (gethostname (jenova_hostname, 128) != 0) {
                        strcpy (jenova_hostname, "(unknown)");
                        log ("warning: couldn't figure out your hostname");
                }
                jenova_hostname[127] = 0;
        }

        if (! jenova_no_daemon) {
                daemonize (PACKAGE);
        }
        signal (SIGTERM, term_signal);

        if (jenova_logfile != NULL) {
                if (! log_open (jenova_logfile)) {
                        fprintf (stderr, "Could not open logfile '%s'!\n", jenova_logfile);
                        exit (1);
                }
        } else {
                if (jenova_no_daemon) {
                        log_set (stderr);
                } else {
                        log_close ();
                }
        }

        log ("----- " PACKAGE "/" VERSION " coming online...");

        /* open up sql connection */
        if (strlen (db_pass) == 0) {
                db_pass = NULL;
        }

        /* should use shared libraries here, or something. */
        if (jenova_gdbm_file != NULL) {
                /* use gdbm */
                db_set_module (client_gdbm_get_module ());
                g_datalist_init (&params);
                g_datalist_set_data (&params, "filename", jenova_gdbm_file);
                global_sql = db_init (params);
                g_datalist_clear (&params);
                if (global_sql == NULL) {
                        log ("ERROR: Couldn't initialize the gdbm database files.");
                        exit (1);
                }
                log ("gdbm started okay");
        } else {
                /* use MySQL */
                db_set_module (client_sql_get_module ());
                g_datalist_init (&params);
                g_datalist_set_data (&params, "host", db_host);
                g_datalist_set_data (&params, "username", db_user);
                g_datalist_set_data (&params, "password", db_pass);
                g_datalist_set_data (&params, "database", db_db);
                global_sql = db_init (params);
                g_datalist_clear (&params);
                if (global_sql == NULL) {
                        log ("ERROR: Couldn't initialize MySQL database connection.");
                        log ("       Possibly you compiled without MySQL support.");
                        exit (1);
                }
        }

        if (jenova_test) {
                Say2Cert *cert;
                GHashTable *table;
                GList *hostlist, *iter;

                table = say2_cert_table_new ();
                if (! say2_cert_table_load (table, "/tmp/foo")) {
                        log ("failed to read table. :(");
                } else {
                        hostlist = say2_cert_table_get_hostnames (table);
                        for (iter = hostlist; iter != NULL; iter = iter->next) {
                                cert = say2_cert_table_lookup (table, (char *)iter->data);
                                log ("host '%s' cert(%d)", (char *)iter->data, cert->length);
                        }
                }
                say2_cert_table_free (table);

                cert = say2_cert_new_host ("kitten.lag.net", 20, "ABCDEFGHIJKLMNOPQRSTUVWXYZ");
                say2_cert_add_alias (cert, "say2.lag.net");
                say2_cert_add_alias (cert, "web.lag.net");

                table = say2_cert_table_new ();
                say2_cert_table_add (table, cert);
                if (! say2_cert_table_save (table, "/tmp/foo")) {
                        log ("failed to write table. :(");
                }

                exit (0);
        }

        if (db_lookup_user != NULL) {
                Client *client;
                char *str;

                client = client_lookup_offline (global_sql, db_lookup_user, CLIENT_USER);
                if (client == NULL) {
                        log ("can't find user '%s'", db_lookup_user);
                        exit (1);
                }

                if (db_set_privs != NULL) {
                        client_string_to_privs (client, db_set_privs);
                        if (! db_update_client (global_sql, client)) {
                                log ("can't update the userdb for some reason");
                        }
                }

                str = client_to_string (client, TRUE);
                log ("%s", str);
                g_free (str);

                client_free (client);
                exit (0);
        }
        if (db_add_user != NULL) {
                int i;
                Client *client;

                if (db_add_email == NULL) {
                        log ("also need --email parameter");
                        exit (1);
                }
                if (db_add_password == NULL) {
                        db_add_password = g_malloc (9);
                        srand (time (NULL));
                        for (i = 0; i < 8; i++) {
                                db_add_password[i] = 'a' + (rand()%26);
                        }
                        db_add_password[8] = '\0';
                }

                client = client_new(FALSE);
                client->id = g_strdup (db_add_user);
                client->email = g_strdup (db_add_email);
                client->flags = CLIENT_USER;

                if (db_set_privs != NULL) {
                        client_string_to_privs (client, db_set_privs);
                }

                if (! db_add_client (global_sql, client)) {
                        log ("failed to add user '%s'", db_add_user);
                        exit (1);
                }
                log ("added user '%s'", db_add_user);
                if (! db_set_password (global_sql, client, db_add_password)) {
                        log ("failed to set password for '%s' though :(", db_add_user);
                        exit (1);
                }
                log ("set password to '%s'", db_add_password);
                client_free(client);
                exit (0);
        }

        listen = socket_new (NULL, jenova_port);
        if (! listen) {
                log ("unable to open port %d!", jenova_port);
                exit (1);
        }
        if (! socket_listen (listen, new_connect)) {
                log ("unable to listen at port %d!", jenova_port);
                exit (1);
        }

        set_keys_on_socket (listen);

        /* okay, drop to a normal user if we're running as root */
        daemon_drop_privs (jenova_run_user);

        buddy_table_init();
        memories_init ();

        /* 5 minute timer for idle checking */
        g_timeout_add (300000, idle_check, NULL);

        uptime = time (NULL);

        log ("Now online & accepting connections.");
	l_main_loop = g_main_new (TRUE);
	while (g_main_is_running (l_main_loop)) {
		g_main_iteration (TRUE);
	}

	g_main_destroy (l_main_loop);
        log ("Shutting down!");
        db_shutdown (global_sql);

        dispatch_free (dispatch_table);
        dispatch_free (login_dispatch_table);
        buddy_table_free();

        exit (0);
}
