/* $Id: say.c,v 1.11 2001/07/04 20:54:46 robey Exp $
 *
 * Channel implementation for say2 server.
 *
 * Copyright (C) 2000  Robey Pointer <robey@lag.net>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <say2/say2.h>
#include "client.h"
#include "say.h"
#include "memory.h"

/* backlog is a list of EnamlBlock* */
static int backlog_max = 200;
static int backlog_size = 0;
static GList *backlog_head = NULL;	/* most recent */
static GList *backlog_tail = NULL;


static void
send_error (Socket *sock, const char *command, const char *msg)
{
        send_enaml_buildf (sock, "error", "command:%s text:%s", command, msg);
}

static void
add_say_log (EnamlBlock *block)
{
        GList *old_tail;

        g_assert (backlog_max >= 1);

        /* all this mucking about just to make GList efficient...  if we have
         * to do all this, is there really any advantage to using GList?
         */
        while (backlog_size >= backlog_max) {
                old_tail = backlog_tail;
                backlog_tail = g_list_previous (backlog_tail);
                enaml_free ((EnamlBlock *) old_tail->data);
                backlog_tail = g_list_remove (backlog_tail, old_tail->data);
                backlog_size--;
        }

        backlog_head = g_list_prepend (backlog_head, enaml_copy_block (block));
        backlog_size++;
        if (backlog_tail == NULL) {
                backlog_tail = backlog_head;
        }
}

void
do_say (Client *client, const EnamlBlock *say, Socket *sock)
{
        EnamlBlock *out;
        EnamlBlock *say_copy;

        if (say == NULL) {
                send_error (sock, "say", "Say requires arguments.");
                return;
        }

        memories_scan (say);

        say_copy = enaml_copy_block (say);
        out = enaml_new ();
        enaml_add_block (out, "say", say_copy, ENAML_FREE);
        enaml_add_string (say_copy, "id", client->id, 0);
        enaml_add_long (say_copy, "time", (long) time (NULL));
        enaml_add_string (say_copy, "nick", client->nick, 0);

        add_say_log (out);
        client_broadcast (out);
        enaml_free (out);

        /* channel chat counts as activity */
        client->activity = time (NULL);
}

void
do_history (Client *client, const EnamlBlock *history, Socket *sock)
{
        GList *iter;
        int limit = backlog_max;
        int count;
        GList *start;

        if (history != NULL) {
                /* might be a limit */
                limit = (int) enaml_get_long (history, "max", limit);
                if (limit < 1) {
                        limit = 1;
                } 
        }

        if (limit < backlog_size) {
                count = limit;
                start = g_list_nth (backlog_head, count-1);
        } else {
                count = backlog_size;
                start = backlog_tail;
        }

        client_sendf (client, "begin-response",
                      "count:%d total:%d response:%s",
                      (long)count, (long)backlog_size, "say");
        for (iter = start; iter != NULL; iter = g_list_previous (iter)) {
                client_send (client, (EnamlBlock *)(iter->data));
        }
        client_sendf (client, "end-response", "response:%s", "say");
}
